@@ -106,6 +106,11 @@
 #endif
 
 static const char *dirname;
+/* whether we have forced fuse in single threaded mode (`-s' option). This happens
+ * because we can't store uid/gids per thread and the file system is accessible for
+ * multiple users via the `-o allow_other' option.
+ */
+static bool single_threaded = false;
 
 void stderr_print(const char *fmt, ...)
 {
@@ -158,16 +163,15 @@ static char *map_path(const char *path)
  * which is used in this implementation.
  */
 
-static size_t get_groups(gid_t **groups)
+static size_t get_groups(pid_t pid, gid_t **groups)
 {
 	static char key[] = "\nGroups:\t";
 	char filename[64], buf[2048], *s, *t, c = '\0';
 	int fd, num_read, matched = 0;
 	size_t n = 0;
 	gid_t *gids, grp = 0;
-	pid_t tid = fuse_get_context()->pid;
 
-	sprintf(filename, "/proc/%u/task/%u/status", tid, tid);
+	sprintf(filename, "/proc/%u/task/%u/status", pid, pid);
 	fd = open(filename, O_RDONLY);
 	if (fd == -1)
 		return 0;
@@ -222,36 +226,75 @@ static size_t get_groups(gid_t **groups)
 	return n;
 }
 
-/* This only works when the fs is mounted by root.
- * Further more it relies on the fact that the euid
- * and egid are stored per thread.
+/* This only works when the filesystem is mounted by root and fuse
+ * operates in single threaded mode. Because the euid/egid are stored
+ * per process this would otherwise cause all sorts of race condidtions
+ * and security issues when multiple users access the file system
+ * simultaneously.
  */
 
-static inline void enter_user_context()
+static inline void enter_user_context_effective()
 {
 	gid_t *groups;
 	size_t ngroups;
 	struct fuse_context *c = fuse_get_context();
 
-	if (getuid() || c->uid == 0)
+	if (!single_threaded || getuid())
 		return;
-	if ((ngroups = get_groups(&groups))) {
+	if ((ngroups = get_groups(c->pid, &groups))) {
 		setgroups(ngroups, groups);
 		free(groups);
 	}
+
 	setegid(c->gid);
 	seteuid(c->uid);
 }
 
-static inline void leave_user_context()
+static inline void leave_user_context_effective()
 {
-	if (getuid())
+	if (!single_threaded || getuid())
 		return;
 
 	seteuid(getuid());
 	setegid(getgid());
 }
 
+/* access(2) checks the real uid/gid not the effective one
+ * we therefore switch them if run as root and in single
+ * threaded mode.
+ *
+ * The real uid/gid are stored per process which is why we
+ * can't change them in multithreaded mode. This would lead
+ * to all sorts of race conditions and security issues when
+ * multiple users access the file system simultaneously.
+ *
+ */
+
+static inline void enter_user_context_real()
+{
+	gid_t *groups;
+	size_t ngroups;
+	struct fuse_context *c = fuse_get_context();
+
+	if (!single_threaded || geteuid())
+		return;
+	if ((ngroups = get_groups(c->pid, &groups))) {
+		setgroups(ngroups, groups);
+		free(groups);
+	}
+	setregid(c->gid, -1);
+	setreuid(c->uid, -1);
+}
+
+static inline void leave_user_context_real()
+{
+	if (!single_threaded || geteuid())
+		return;
+
+	setuid(geteuid());
+	setgid(getegid());
+}
+
 static ssize_t ciopfs_get_orig_name(const char *path, char *value, size_t size)
 {
 	ssize_t attrlen;
@@ -314,7 +357,9 @@ static int ciopfs_getattr(const char *path, struct stat *st_data)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
+	enter_user_context_effective();
 	int res = lstat(p, st_data);
+	leave_user_context_effective();
 	free(p);
 	return (res == -1) ? -errno : 0;
 }
@@ -322,21 +367,22 @@ static int ciopfs_getattr(const char *path, struct stat *st_data)
 static int ciopfs_fgetattr(const char *path, struct stat *stbuf,
                            struct fuse_file_info *fi)
 {
+	enter_user_context_effective();
 	int res = fstat(fi->fh, stbuf);
+	leave_user_context_effective();
 	if (res == -1)
 		return -errno;
 	return 0;
 }
 
-
 static int ciopfs_readlink(const char *path, char *buf, size_t size)
 {
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = readlink(p, buf, size - 1);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -344,11 +390,28 @@ static int ciopfs_readlink(const char *path, char *buf, size_t size)
 	return 0;
 }
 
+static int ciopfs_opendir(const char *path, struct fuse_file_info *fi)
+{
+	char *p = map_path(path);
+	if (unlikely(p == NULL))
+		return -ENOMEM;
+	enter_user_context_effective();
+	DIR *dp = opendir(p);
+	leave_user_context_effective();
+	free(p);
+
+	if (dp == NULL)
+		return -errno;
+
+	fi->fh = (uint64_t)(uintptr_t)dp;
+	return 0;
+}
+
 static int ciopfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                           off_t offset, struct fuse_file_info *fi)
 {
 	int ret = 0;
-	DIR *dp;
+	DIR *dp = (DIR *)(uintptr_t)fi->fh;
 	struct dirent *de;
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
@@ -362,17 +425,13 @@ static int ciopfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
 		goto out;
 	}
 
-	strcpy(dnamebuf, p);
-
-	(void) offset;
-	(void) fi;
-
-	dp = opendir(p);
-	if (dp == NULL) {
-		ret = -errno;
+	if (!dp) {
+		ret = -EBADF;
 		goto out;
 	}
 
+	seekdir(dp, offset);
+
 	while ((de = readdir(dp)) != NULL) {
 		struct stat st;
 		char *dname;
@@ -411,23 +470,29 @@ static int ciopfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
 				dname = de->d_name;
 		}
 		debug("dname: %s\n", dname);
-		if (filler(buf, dname, &st, 0))
+		if (filler(buf, dname, &st, telldir(dp)))
 			break;
 	}
 
-	closedir(dp);
 out:
 	free(p);
 	return ret;
 }
 
+static int ciopfs_releasedir(const char *path, struct fuse_file_info *fi)
+{
+	if (fi->fh)
+		closedir((DIR *)(uintptr_t)fi->fh);
+	return 0;
+}
+
 static int ciopfs_mknod(const char *path, mode_t mode, dev_t rdev)
 {
 	int res;
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	/* On Linux this could just be 'mknod(p, mode, rdev)' but this
 	   is more portable */
 	if (S_ISREG(mode)) {
@@ -440,7 +505,7 @@ static int ciopfs_mknod(const char *path, mode_t mode, dev_t rdev)
 		res = mkfifo(p, mode);
 	} else
 		res = mknod(p, mode, rdev);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -454,9 +519,9 @@ static int ciopfs_mkdir(const char *path, mode_t mode)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = mkdir(p, mode);
-	leave_user_context();
+	leave_user_context_effective();
 
 	if (res == -1) {
 		ret =  -errno;
@@ -474,9 +539,9 @@ static int ciopfs_unlink(const char *path)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = unlink(p);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -488,9 +553,9 @@ static int ciopfs_rmdir(const char *path)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = rmdir(p);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -504,9 +569,9 @@ static int ciopfs_symlink(const char *from, const char *to)
 	char *t = map_path(to);
 	if (unlikely(f == NULL || t == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = symlink(f, t);
-	leave_user_context();
+	leave_user_context_effective();
 	if (res == -1) {
 		ret = -errno;
 		goto out;
@@ -525,9 +590,9 @@ static int ciopfs_rename(const char *from, const char *to)
 	char *t = map_path(to);
 	if (unlikely(f == NULL || t == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = rename(f, t);
-	leave_user_context();
+	leave_user_context_effective();
 	if (res == -1) {
 		ret = -errno;
 		goto out;
@@ -546,9 +611,9 @@ static int ciopfs_link(const char *from, const char *to)
 	char *t = map_path(to);
 	if (unlikely(f == NULL || t == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = link(f, t);
-	leave_user_context();
+	leave_user_context_effective();
 	if (res == -1) {
 		ret = -errno;
 		goto out;
@@ -565,9 +630,9 @@ static int ciopfs_chmod(const char *path, mode_t mode)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = chmod(p, mode);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -579,9 +644,9 @@ static int ciopfs_chown(const char *path, uid_t uid, gid_t gid)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = lchown(p, uid, gid);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -593,9 +658,9 @@ static int ciopfs_truncate(const char *path, off_t size)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = truncate(p, size);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -604,9 +669,9 @@ static int ciopfs_truncate(const char *path, off_t size)
 
 static int ciopfs_ftruncate(const char *path, off_t size, struct fuse_file_info *fi)
 {
-	enter_user_context();
+	enter_user_context_effective();
 	int res = ftruncate(fi->fh, size);
-	leave_user_context();
+	leave_user_context_effective();
 	if (res == -1)
 		return -errno;
 
@@ -625,9 +690,9 @@ static int ciopfs_utimens(const char *path, const struct timespec ts[2])
 	tv[1].tv_sec = ts[1].tv_sec;
 	tv[1].tv_usec = ts[1].tv_nsec / 1000;
 
-	enter_user_context();
+	enter_user_context_effective();
 	int res = utimes(p, tv);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -639,9 +704,9 @@ static int ciopfs_create(const char *path, mode_t mode, struct fuse_file_info *f
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int fd = open(p, fi->flags, mode);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (fd == -1)
 		return -errno;
@@ -655,9 +720,9 @@ static int ciopfs_open(const char *path, struct fuse_file_info *fi)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int fd = open(p, fi->flags);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (fd == -1)
 		return -errno;
@@ -690,9 +755,9 @@ static int ciopfs_statfs(const char *path, struct statvfs *stbuf)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = statvfs(p, stbuf);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -739,9 +804,9 @@ static int ciopfs_access(const char *path, int mode)
   	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_real();
   	int res = access(p, mode);
-	leave_user_context();
+	leave_user_context_real();
 	free(p);
   	if (res == -1)
     		return -errno;
@@ -758,9 +823,9 @@ static int ciopfs_setxattr(const char *path, const char *name, const char *value
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = lsetxattr(p, name, value, size, flags);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -772,9 +837,9 @@ static int ciopfs_getxattr(const char *path, const char *name, char *value, size
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = lgetxattr(p, name, value, size);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -786,9 +851,9 @@ static int ciopfs_listxattr(const char *path, char *list, size_t size)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = llistxattr(p, list, size);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -804,9 +869,9 @@ static int ciopfs_removexattr(const char *path, const char *name)
 	char *p = map_path(path);
 	if (unlikely(p == NULL))
 		return -ENOMEM;
-	enter_user_context();
+	enter_user_context_effective();
 	int res = lremovexattr(p, name);
-	leave_user_context();
+	leave_user_context_effective();
 	free(p);
 	if (res == -1)
 		return -errno;
@@ -833,7 +898,9 @@ struct fuse_operations ciopfs_operations = {
 	.getattr	= ciopfs_getattr,
 	.fgetattr	= ciopfs_fgetattr,
 	.readlink	= ciopfs_readlink,
+	.opendir	= ciopfs_opendir,
 	.readdir	= ciopfs_readdir,
+	.releasedir	= ciopfs_releasedir,
 	.mknod		= ciopfs_mknod,
 	.mkdir		= ciopfs_mkdir,
 	.symlink	= ciopfs_symlink,
@@ -861,12 +928,6 @@ struct fuse_operations ciopfs_operations = {
 	.removexattr	= ciopfs_removexattr,
 	.lock		= ciopfs_lock,
 	.init		= ciopfs_init
-/*
- *	what about:
- *
- *	opendir
- *	releasedir
- */
 };
 
 static void usage(const char *name)
@@ -908,6 +969,14 @@ static int ciopfs_opt_parse(void *data, const char *arg, int key, struct fuse_ar
 					case 'f':
 						dolog = stderr_print;
 				}
+			} else if (!strcmp("allow_other", arg)) {
+				/* disable multithreaded mode if the file system
+				 * is accessible to multiple users simultanousely
+				 * because we can't store uid/gid per thread and
+				 * this leads to all sorts of race conditions and
+				 * security issues.
+				 */
+				single_threaded = (getuid() == 0);
 			}
 			return 1;
 		case CIOPFS_OPT_HELP:
@@ -935,10 +1004,15 @@ static struct fuse_opt ciopfs_opts[] = {
 int main(int argc, char *argv[])
 {
 	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
-	if (fuse_opt_parse(&args, &dirname, ciopfs_opts, ciopfs_opt_parse)) {
-		fprintf(stderr, "Invalid arguments, see `%s -h' for usage\n", argv[0]);
-		exit(1);
+	fuse_opt_parse(&args, &dirname, ciopfs_opts, ciopfs_opt_parse);
+
+	if (single_threaded) {
+		fuse_opt_add_arg(&args, "-s");
+		log_print("disabling multithreaded mode for root mounted "
+		          "filesystem that is accessible for other users "
+		          "via the `-o allow_other' option\n");
 	}
+
 	umask(0);
 	return fuse_main(args.argc, args.argv, &ciopfs_operations, NULL);
 }
